@@ -20,7 +20,6 @@ type compiler struct {
 	codeinfos     []codeinfo
 	scopes        []*scopeinfo
 	scopecnt      int
-	funcs         []*funcinfo
 }
 
 // Code is a compiled jq query.
@@ -68,9 +67,10 @@ type codeinfo struct {
 }
 
 type scopeinfo struct {
+	variables   []*varinfo
+	funcs       []*funcinfo
 	id          int
 	depth       int
-	variables   []varinfo
 	variablecnt int
 }
 
@@ -81,10 +81,9 @@ type varinfo struct {
 }
 
 type funcinfo struct {
-	name      string
-	pc        int
-	args      []string
-	argsorder []int
+	name   string
+	pc     int
+	argcnt int
 }
 
 // Compile compiles a query.
@@ -95,9 +94,9 @@ func Compile(q *Query, options ...CompilerOption) (*Code, error) {
 	}
 	scope := c.newScope()
 	c.scopes = []*scopeinfo{scope}
-	defer c.lazy(func() *code {
-		return &code{op: opscope, v: [2]int{scope.id, scope.variablecnt}}
-	})()
+	setscope := c.lazy(func() *code {
+		return &code{op: opscope, v: [3]int{scope.id, scope.variablecnt, 0}}
+	})
 	if c.moduleLoader != nil {
 		if moduleLoader, ok := c.moduleLoader.(interface {
 			LoadInitModules() ([]*Query, error)
@@ -116,6 +115,7 @@ func Compile(q *Query, options ...CompilerOption) (*Code, error) {
 	if err := c.compile(q); err != nil {
 		return nil, err
 	}
+	setscope()
 	c.optimizeTailRec()
 	c.optimizeJumps()
 	return &Code{
@@ -130,8 +130,8 @@ func (c *compiler) compile(q *Query) error {
 		if !newLexer(name).validVarName() {
 			return &variableNameError{name}
 		}
-		v := c.pushVariable(name)
-		c.append(&code{op: opstore, v: v})
+		c.appendCodeInfo(name)
+		c.append(&code{op: opstore, v: c.pushVariable(name)})
 	}
 	for _, i := range q.Imports {
 		if err := c.compileImport(i); err != nil {
@@ -208,10 +208,10 @@ func (c *compiler) compileModule(q *Query, alias string) error {
 	}(len(scope.variables))
 	if alias != "" {
 		defer func(l int) {
-			for _, f := range c.funcs[l:] {
+			for _, f := range scope.funcs[l:] {
 				f.name = alias + "::" + f.name
 			}
-		}(len(c.funcs))
+		}(len(scope.funcs))
 	}
 	for _, i := range q.Imports {
 		if err := c.compileImport(i); err != nil {
@@ -227,106 +227,134 @@ func (c *compiler) compileModule(q *Query, alias string) error {
 }
 
 func (c *compiler) newVariable() [2]int {
-	return c.pushVariable("")
+	return c.createVariable("")
 }
 
 func (c *compiler) pushVariable(name string) [2]int {
 	s := c.scopes[len(c.scopes)-1]
-	if name != "" {
-		for _, v := range s.variables {
-			if v.name == name && v.depth == s.depth {
-				return v.index
+	for _, v := range s.variables {
+		if v.name == name && v.depth == s.depth {
+			return v.index
+		}
+	}
+	return c.createVariable(name)
+}
+
+func (c *compiler) createVariable(name string) [2]int {
+	s := c.scopes[len(c.scopes)-1]
+	v := [2]int{s.id, s.variablecnt}
+	s.variablecnt++
+	s.variables = append(s.variables, &varinfo{name, v, s.depth})
+	return v
+}
+
+func (c *compiler) lookupVariable(name string) ([2]int, error) {
+	for i := len(c.scopes) - 1; i >= 0; i-- {
+		s := c.scopes[i]
+		for j := len(s.variables) - 1; j >= 0; j-- {
+			if w := s.variables[j]; w.name == name {
+				return w.index, nil
 			}
 		}
 	}
-	v := [2]int{s.id, s.variablecnt}
-	s.variablecnt++
-	s.variables = append(s.variables, varinfo{name, v, s.depth})
-	return v
+	return [2]int{}, &variableNotFoundError{name}
+}
+
+func (c *compiler) lookupFuncOrVariable(name string) (*funcinfo, *varinfo) {
+	for i, isFunc := len(c.scopes)-1, name[0] != '$'; i >= 0; i-- {
+		s := c.scopes[i]
+		if isFunc {
+			for j := len(s.funcs) - 1; j >= 0; j-- {
+				if f := s.funcs[j]; f.name == name && f.argcnt == 0 {
+					return f, nil
+				}
+			}
+		}
+		for j := len(s.variables) - 1; j >= 0; j-- {
+			if v := s.variables[j]; v.name == name {
+				return nil, v
+			}
+		}
+	}
+	return nil, nil
 }
 
 func (c *compiler) newScope() *scopeinfo {
 	i := c.scopecnt // do not use len(c.scopes) because it pops
 	c.scopecnt++
-	return &scopeinfo{i, 0, nil, 0}
+	return &scopeinfo{id: i}
 }
 
 func (c *compiler) newScopeDepth() func() {
 	scope := c.scopes[len(c.scopes)-1]
-	l, m := len(scope.variables), len(c.funcs)
+	l, m := len(scope.variables), len(scope.funcs)
 	scope.depth++
 	return func() {
 		scope.depth--
 		scope.variables = scope.variables[:l]
-		c.funcs = c.funcs[:m]
+		scope.funcs = scope.funcs[:m]
 	}
 }
 
 func (c *compiler) compileFuncDef(e *FuncDef, builtin bool) error {
+	var scope *scopeinfo
 	if builtin {
-		for i := len(c.funcs) - 1; i >= 0; i-- {
-			if f := c.funcs[i]; f.name == e.Name && len(f.args) == len(e.Args) {
+		scope = c.scopes[0]
+		for i := len(scope.funcs) - 1; i >= 0; i-- {
+			if f := scope.funcs[i]; f.name == e.Name && f.argcnt == len(e.Args) {
 				return nil
 			}
 		}
+	} else {
+		scope = c.scopes[len(c.scopes)-1]
 	}
 	defer c.lazy(func() *code {
 		return &code{op: opjump, v: c.pc()}
 	})()
 	c.appendCodeInfo(e.Name)
 	defer c.appendCodeInfo("end of " + e.Name)
-	pc, argsorder := c.pc(), getArgsOrder(e.Args)
-	c.funcs = append(c.funcs, &funcinfo{e.Name, pc, e.Args, argsorder})
-	defer func(l, m int, variables []string) {
-		c.scopes, c.funcs, c.variables = c.scopes[:l], c.funcs[:m], variables
-	}(len(c.scopes), len(c.funcs), c.variables)
+	pc := c.pc()
+	scope.funcs = append(scope.funcs, &funcinfo{e.Name, pc, len(e.Args)})
+	defer func(l int, variables []string) {
+		c.scopes, c.variables = c.scopes[:l], variables
+	}(len(c.scopes), c.variables)
 	c.variables = c.variables[len(c.variables):]
-	scope := c.newScope()
+	scope = c.newScope()
 	c.scopes = append(c.scopes, scope)
 	defer c.lazy(func() *code {
-		return &code{op: opscope, v: [2]int{scope.id, scope.variablecnt}}
+		return &code{op: opscope, v: [3]int{scope.id, scope.variablecnt, len(e.Args)}}
 	})()
 	if len(e.Args) > 0 {
+		type varIndex struct {
+			name  string
+			index [2]int
+		}
+		vis := make([]varIndex, 0, len(e.Args))
 		v := c.newVariable()
 		c.append(&code{op: opstore, v: v})
-		skip := make([]bool, len(e.Args))
-		for i, name := range e.Args {
-			for j := 0; j < i; j++ {
-				if name == e.Args[j] {
-					skip[j] = true
-					break
-				}
+		for _, arg := range e.Args {
+			if arg[0] == '$' {
+				c.appendCodeInfo(arg[1:])
+				w := c.createVariable(arg[1:])
+				c.append(&code{op: opstore, v: w})
+				vis = append(vis, varIndex{arg, w})
+			} else {
+				c.appendCodeInfo(arg)
+				c.append(&code{op: opstore, v: c.createVariable(arg)})
 			}
 		}
-		for _, i := range argsorder {
-			if skip[i] {
-				c.append(&code{op: oppop})
-			} else {
-				c.append(&code{op: opstore, v: c.pushVariable(e.Args[i])})
-			}
+		for _, w := range vis {
+			c.append(&code{op: opload, v: v})
+			c.append(&code{op: opload, v: w.index})
+			c.append(&code{op: opcallpc})
+			c.appendCodeInfo(w.name)
+			c.append(&code{op: opstore, v: c.pushVariable(w.name)})
 		}
 		c.append(&code{op: opload, v: v})
 	}
 	return c.compile(e.Body)
 }
 
-func getArgsOrder(args []string) []int {
-	xs := make([]int, len(args))
-	if len(xs) > 0 {
-		for i := range xs {
-			xs[i] = i
-		}
-		sort.Slice(xs, func(i, j int) bool {
-			xi, xj := xs[i], xs[j]
-			if args[xi][0] == '$' {
-				return args[xj][0] == '$' && xi > xj // reverse the order of variables
-			}
-			return args[xj][0] == '$' || xi < xj
-		})
-	}
-	return xs
-}
-
 func (c *compiler) compileQuery(e *Query) error {
 	for _, fd := range e.FuncDefs {
 		if err := c.compileFuncDef(fd, false); err != nil {
@@ -521,11 +549,10 @@ func (c *compiler) compileBind(b *Bind) error {
 	if len(b.Patterns) > 1 {
 		pc = c.pc()
 	}
-	if c.codes[len(c.codes)-2].op == opexpbegin {
-		c.codes[len(c.codes)-2] = c.codes[len(c.codes)-1]
-		c.codes = c.codes[:len(c.codes)-1]
+	if len(b.Patterns) == 1 && c.codes[len(c.codes)-2].op == opexpbegin {
+		c.codes[len(c.codes)-2].op = opnop
 	} else {
-		c.append(&code{op: opexpend}) // ref: compileQuery
+		c.append(&code{op: opexpend}) // ref: compileTermSuffix
 	}
 	return c.compileQuery(b.Body)
 }
@@ -611,12 +638,18 @@ func (c *compiler) compileIf(e *If) error {
 	c.appendCodeInfo(e)
 	c.append(&code{op: opdup}) // duplicate the value for then or else clause
 	c.append(&code{op: opexpbegin})
+	pc := len(c.codes)
 	f := c.newScopeDepth()
 	if err := c.compileQuery(e.Cond); err != nil {
 		return err
 	}
 	f()
-	c.append(&code{op: opexpend})
+	if pc == len(c.codes) {
+		c.codes = c.codes[:pc-1]
+	} else {
+		c.append(&code{op: opexpend})
+	}
+	pcc := len(c.codes)
 	setjumpifnot := c.lazy(func() *code {
 		return &code{op: opjumpifnot, v: c.pc() + 1} // if falsy, skip then clause
 	})
@@ -634,6 +667,16 @@ func (c *compiler) compileIf(e *If) error {
 	}
 	if e.Else != nil {
 		defer c.newScopeDepth()()
+		defer func() {
+			// optimize constant results
+			//    opdup, ..., opjumpifnot, opconst, opjump, opconst
+			// => opnop, ..., opjumpifnot, oppush,  opjump, oppush
+			if pcc+4 == len(c.codes) && c.codes[pcc+1].op == opconst && c.codes[pcc+3].op == opconst {
+				c.codes[pc-2].op = opnop
+				c.codes[pcc+1].op = oppush
+				c.codes[pcc+3].op = oppush
+			}
+		}()
 		return c.compileQuery(e.Else)
 	}
 	return nil
@@ -729,12 +772,30 @@ func (c *compiler) compileForeach(e *Foreach) error {
 
 func (c *compiler) compileLabel(e *Label) error {
 	c.appendCodeInfo(e)
+	v := c.pushVariable("$%" + e.Ident[1:])
 	defer c.lazy(func() *code {
-		return &code{op: opforklabel, v: e.Ident}
+		return &code{op: opforklabel, v: v}
 	})()
 	return c.compileQuery(e.Body)
 }
 
+func (c *compiler) compileBreak(label string) error {
+	v, err := c.lookupVariable("$%" + label[1:])
+	if err != nil {
+		return &breakError{label, nil}
+	}
+	c.append(&code{op: oppop})
+	c.append(&code{op: opload, v: v})
+	c.append(&code{op: opcall, v: [3]interface{}{
+		func(v interface{}, _ []interface{}) interface{} {
+			return &breakError{label, v}
+		},
+		0,
+		"_break",
+	}})
+	return nil
+}
+
 func (c *compiler) compileTerm(e *Term) error {
 	if len(e.SuffixList) > 0 {
 		s := e.SuffixList[len(e.SuffixList)-1]
@@ -788,8 +849,7 @@ func (c *compiler) compileTerm(e *Term) error {
 	case TermTypeLabel:
 		return c.compileLabel(e.Label)
 	case TermTypeBreak:
-		c.append(&code{op: opconst, v: e.Break})
-		return c.compileCall("_break", nil)
+		return c.compileBreak(e.Break)
 	case TermTypeQuery:
 		defer c.newScopeDepth()()
 		return c.compileQuery(e.Query)
@@ -819,42 +879,43 @@ func (c *compiler) compileIndex(e *Term, x *Index) error {
 }
 
 func (c *compiler) compileFunc(e *Func) error {
-	for i := len(c.funcs) - 1; i >= 0; i-- {
-		if f := c.funcs[i]; f.name == e.Name && len(f.args) == len(e.Args) {
-			return c.compileCallPc(f, e.Args)
-		}
-	}
-	for i := len(c.scopes) - 1; i >= 0; i-- {
-		s := c.scopes[i]
-		for j := len(s.variables) - 1; j >= 0; j-- {
-			v := s.variables[j]
-			if v.name == e.Name && len(e.Args) == 0 {
-				if e.Name[0] == '$' {
-					c.append(&code{op: oppop})
-					c.append(&code{op: opload, v: v.index})
-				} else {
-					c.append(&code{op: opload, v: v.index})
-					c.append(&code{op: opcallpc})
-				}
-				return nil
-			}
-		}
-	}
-	if (e.Name == "$ENV" || e.Name == "env") && len(e.Args) == 0 {
-		env := make(map[string]interface{})
-		if c.environLoader != nil {
-			for _, kv := range c.environLoader() {
-				xs := strings.SplitN(kv, "=", 2)
-				env[xs[0]] = xs[1]
-			}
-		}
-		c.append(&code{op: opconst, v: env})
-		return nil
-	}
-	if e.Name[0] == '$' {
-		return &variableNotFoundError{e.Name}
-	}
 	name := e.Name
+	if len(e.Args) == 0 {
+		if f, v := c.lookupFuncOrVariable(name); f != nil {
+			return c.compileCallPc(f, e.Args)
+		} else if v != nil {
+			if name[0] == '$' {
+				c.append(&code{op: oppop})
+				c.append(&code{op: opload, v: v.index})
+			} else {
+				c.append(&code{op: opload, v: v.index})
+				c.append(&code{op: opcallpc})
+			}
+			return nil
+		} else if name == "$ENV" || name == "env" {
+			env := make(map[string]interface{})
+			if c.environLoader != nil {
+				for _, kv := range c.environLoader() {
+					if i := strings.IndexByte(kv, '='); i > 0 {
+						env[kv[:i]] = kv[i+1:]
+					}
+				}
+			}
+			c.append(&code{op: opconst, v: env})
+			return nil
+		} else if name[0] == '$' {
+			return &variableNotFoundError{name}
+		}
+	} else {
+		for i := len(c.scopes) - 1; i >= 0; i-- {
+			s := c.scopes[i]
+			for j := len(s.funcs) - 1; j >= 0; j-- {
+				if f := s.funcs[j]; f.name == name && f.argcnt == len(e.Args) {
+					return c.compileCallPc(f, e.Args)
+				}
+			}
+		}
+	}
 	if name[0] == '_' {
 		name = name[1:]
 	}
@@ -866,8 +927,9 @@ func (c *compiler) compileFunc(e *Func) error {
 				}
 			}
 		}
-		for i := len(c.funcs) - 1; i >= 0; i-- {
-			if f := c.funcs[i]; f.name == e.Name && len(f.args) == len(e.Args) {
+		s := c.scopes[0]
+		for i := len(s.funcs) - 1; i >= 0; i-- {
+			if f := s.funcs[i]; f.name == e.Name && f.argcnt == len(e.Args) {
 				return c.compileCallPc(f, e.Args)
 			}
 		}
@@ -888,7 +950,7 @@ func (c *compiler) compileFunc(e *Func) error {
 			return c.compileCallInternal(
 				[3]interface{}{c.funcBuiltins, 0, e.Name},
 				e.Args,
-				nil,
+				true,
 				false,
 			)
 		case "input":
@@ -898,14 +960,14 @@ func (c *compiler) compileFunc(e *Func) error {
 			return c.compileCallInternal(
 				[3]interface{}{c.funcInput, 0, e.Name},
 				e.Args,
-				nil,
+				true,
 				false,
 			)
 		case "modulemeta":
 			return c.compileCallInternal(
 				[3]interface{}{c.funcModulemeta, 0, e.Name},
 				e.Args,
-				nil,
+				true,
 				false,
 			)
 		default:
@@ -916,7 +978,7 @@ func (c *compiler) compileFunc(e *Func) error {
 		if err := c.compileCallInternal(
 			[3]interface{}{fn.callback, len(e.Args), e.Name},
 			e.Args,
-			nil,
+			true,
 			false,
 		); err != nil {
 			return err
@@ -1140,10 +1202,10 @@ func (c *compiler) compileArray(e *Array) error {
 	arr := c.newVariable()
 	c.append(&code{op: opstore, v: arr})
 	pc := len(c.codes)
-	c.append(&code{op: opnop})
+	c.append(&code{op: opfork})
 	defer func() {
 		if pc < len(c.codes) {
-			c.codes[pc] = &code{op: opfork, v: c.pc() - 2}
+			c.codes[pc].v = c.pc() - 2
 		}
 	}()
 	defer c.newScopeDepth()()
@@ -1158,12 +1220,12 @@ func (c *compiler) compileArray(e *Array) error {
 		return nil
 	}
 	// optimize constant arrays
-	l := e.Query.countCommaQueries()
-	if 3*l != len(c.codes)-pc-3 {
+	if (len(c.codes)-pc)%3 != 0 {
 		return nil
 	}
+	l := (len(c.codes) - pc - 3) / 3
 	for i := 0; i < l; i++ {
-		if (i > 0 && c.codes[pc+i].op != opfork) ||
+		if c.codes[pc+i].op != opfork ||
 			c.codes[pc+i*2+l].op != opconst ||
 			(i < l-1 && c.codes[pc+i*2+l+1].op != opjump) {
 			return nil
@@ -1240,18 +1302,18 @@ func (c *compiler) compileString(s *String, f *Func) error {
 	if f == nil {
 		f = &Func{Name: "tostring"}
 	}
-	e := s.Queries[0]
-	if e.Term.Str == nil {
-		e = &Query{Left: e, Op: OpPipe, Right: &Query{Term: &Term{Type: TermTypeFunc, Func: f}}}
-	}
-	for i := 1; i < len(s.Queries); i++ {
-		x := s.Queries[i]
-		if x.Term.Str == nil {
-			x = &Query{Left: x, Op: OpPipe, Right: &Query{Term: &Term{Type: TermTypeFunc, Func: f}}}
+	var q *Query
+	for _, e := range s.Queries {
+		if e.Term.Str == nil {
+			e = &Query{Left: e, Op: OpPipe, Right: &Query{Term: &Term{Type: TermTypeFunc, Func: f}}}
+		}
+		if q == nil {
+			q = e
+		} else {
+			q = &Query{Left: q, Op: OpAdd, Right: e}
 		}
-		e = &Query{Left: e, Op: OpAdd, Right: x}
 	}
-	return c.compileQuery(e)
+	return c.compileQuery(q)
 }
 
 func (c *compiler) compileTermSuffix(e *Term, s *Suffix) error {
@@ -1291,26 +1353,17 @@ func (c *compiler) compileCall(name string, args []*Query) error {
 	return c.compileCallInternal(
 		[3]interface{}{internalFuncs[name].callback, len(args), name},
 		args,
-		nil,
+		true,
 		name == "_index" || name == "_slice",
 	)
 }
 
 func (c *compiler) compileCallPc(fn *funcinfo, args []*Query) error {
-	if len(args) == 0 {
-		return c.compileCallInternal(fn.pc, args, nil, false)
-	}
-	xs, vars := make([]*Query, len(args)), make(map[int]bool, len(fn.args))
-	for i, j := range fn.argsorder {
-		xs[i] = args[j]
-		if fn.args[j][0] == '$' {
-			vars[i] = true
-		}
-	}
-	return c.compileCallInternal(fn.pc, xs, vars, false)
+	return c.compileCallInternal(fn.pc, args, false, false)
 }
 
-func (c *compiler) compileCallInternal(fn interface{}, args []*Query, vars map[int]bool, indexing bool) error {
+func (c *compiler) compileCallInternal(
+	fn interface{}, args []*Query, internal, indexing bool) error {
 	if len(args) == 0 {
 		c.append(&code{op: opcall, v: fn})
 		return nil
@@ -1326,13 +1379,14 @@ func (c *compiler) compileCallInternal(fn interface{}, args []*Query, vars map[i
 		if err := c.compileFuncDef(&FuncDef{Name: name, Body: args[i]}, false); err != nil {
 			return err
 		}
-		if vars == nil || vars[i] {
+		if internal {
 			switch c.pc() - pc {
 			case 2: // optimize identity argument (opscope, opret)
 				j := len(c.codes) - 3
 				c.codes[j] = &code{op: opload, v: idx}
 				c.codes = c.codes[:j+1]
-				c.funcs = c.funcs[:len(c.funcs)-1]
+				s := c.scopes[len(c.scopes)-1]
+				s.funcs = s.funcs[:len(s.funcs)-1]
 				c.deleteCodeInfo(name)
 			case 3: // optimize one instruction argument (opscope, opX, opret)
 				j := len(c.codes) - 4
@@ -1344,7 +1398,8 @@ func (c *compiler) compileCallInternal(fn interface{}, args []*Query, vars map[i
 					c.codes[j+1] = c.codes[j+2]
 					c.codes = c.codes[:j+2]
 				}
-				c.funcs = c.funcs[:len(c.funcs)-1]
+				s := c.scopes[len(c.scopes)-1]
+				s.funcs = s.funcs[:len(s.funcs)-1]
 				c.deleteCodeInfo(name)
 			default:
 				c.append(&code{op: opload, v: idx})
@@ -1378,24 +1433,27 @@ func (c *compiler) pc() int {
 
 func (c *compiler) lazy(f func() *code) func() {
 	i := len(c.codes)
-	c.codes = append(c.codes, &code{op: opnop})
+	c.codes = append(c.codes, nil)
 	return func() { c.codes[i] = f() }
 }
 
 func (c *compiler) optimizeTailRec() {
 	var pcs []int
-	targets := map[int]bool{}
+	scopes := map[int]bool{}
 L:
 	for i, l := 0, len(c.codes); i < l; i++ {
 		switch c.codes[i].op {
 		case opscope:
 			pcs = append(pcs, i)
-			if c.codes[i].v.([2]int)[1] == 0 {
-				targets[i] = true
+			if v := c.codes[i].v.([3]int); v[2] == 0 {
+				scopes[i] = v[1] == 0
 			}
 		case opcall:
+			var canjump bool
 			if j, ok := c.codes[i].v.(int); !ok ||
-				len(pcs) == 0 || pcs[len(pcs)-1] != j || !targets[j] {
+				len(pcs) == 0 || pcs[len(pcs)-1] != j {
+				break
+			} else if canjump, ok = scopes[j]; !ok {
 				break
 			}
 			for j := i + 1; j < l; {
@@ -1403,7 +1461,12 @@ L:
 				case opjump:
 					j = c.codes[j].v.(int)
 				case opret:
-					c.codes[i] = &code{op: opjump, v: pcs[len(pcs)-1] + 1}
+					if canjump {
+						c.codes[i].op = opjump
+						c.codes[i].v = pcs[len(pcs)-1] + 1
+					} else {
+						c.codes[i].op = opcallrec
+					}
 					continue L
 				default:
 					continue L
